@@ -1,43 +1,14 @@
 from pathlib import Path
 import shutil
 import tempfile
-import traceback
-
-from nbclient import execute as executenb
 
 # from jupyter_client.kernelspec import get_kernel_spec, NoSuchKernel
 
 from jupyter_cache.executors.base import JupyterExecutorAbstract
+from jupyter_cache.executors.utils import single_nb_execution
 from jupyter_cache.cache.main import NbBundleIn, NbArtifacts
 from jupyter_cache.cache.db import NbStageRecord
-from jupyter_cache.utils import to_relative_paths, Timer
-
-
-def copy_assets(record, folder):
-    """Copy notebook assets to the folder the notebook will be executed in."""
-    asset_files = []
-    relative_paths = to_relative_paths(record.assets, Path(record.uri).parent)
-    for path, rel_path in zip(record.assets, relative_paths):
-        temp_file = Path(folder).joinpath(rel_path)
-        temp_file.parent.mkdir(parents=True, exist_ok=True)
-        shutil.copyfile(path, temp_file)
-        asset_files.append(temp_file)
-    return asset_files
-
-
-def create_bundle(nb_bundle, tmpdirname, asset_files, exec_time, exec_tb=None):
-    """Create a cache bundle."""
-    return NbBundleIn(
-        nb_bundle.nb,
-        nb_bundle.uri,
-        # TODO retrieve assets that have changed file mtime?
-        artifacts=NbArtifacts(
-            [p for p in Path(tmpdirname).glob("**/*") if p not in asset_files],
-            tmpdirname,
-        ),
-        data={"execution_seconds": exec_time},
-        traceback=exec_tb,
-    )
+from jupyter_cache.utils import to_relative_paths
 
 
 class ExecutionError(Exception):
@@ -58,7 +29,13 @@ class JupyterExecutorBasic(JupyterExecutorAbstract):
     """
 
     def run_and_cache(
-        self, filter_uris=None, filter_pks=None, converter=None, timeout=30
+        self,
+        filter_uris=None,
+        filter_pks=None,
+        converter=None,
+        timeout=30,
+        allow_errors=False,
+        run_in_temp=True,
     ):
         """This function interfaces with the cache, deferring execution to `execute`."""
         # Get the notebook tha require re-execution
@@ -93,7 +70,9 @@ class JupyterExecutorBasic(JupyterExecutorAbstract):
                     yield stage_record, nb_bundle
 
         # The execute method yields notebook bundles, or ExecutionError
-        for bundle_or_exc in self.execute(_iterator(), int(timeout)):
+        for bundle_or_exc in self.execute(
+            _iterator(), int(timeout), allow_errors, run_in_temp
+        ):
             if isinstance(bundle_or_exc, ExecutionError):
                 self.logger.error(bundle_or_exc.uri, exc_info=bundle_or_exc.exc)
                 result["errored"].append(bundle_or_exc.uri)
@@ -130,7 +109,7 @@ class JupyterExecutorBasic(JupyterExecutorAbstract):
 
         return result
 
-    def execute(self, input_iterator, timeout=30):
+    def execute(self, input_iterator, timeout=30, allow_errors=False, in_temp=True):
         """This function is isolated from the cache, and is responsible for execution.
 
         The method is only supplied with the staged record and input notebook bundle,
@@ -141,44 +120,74 @@ class JupyterExecutorBasic(JupyterExecutorAbstract):
                 uri = nb_bundle.uri
                 self.logger.info("Executing: {}".format(uri))
 
-                with tempfile.TemporaryDirectory() as tmpdirname:
-                    try:
-                        asset_files = copy_assets(stage_record, tmpdirname)
-                    except Exception as err:
-                        yield ExecutionError("Assets Retrieval Error", uri, err)
-                        continue
-                    timer = Timer()
-                    try:
-                        with timer:
-                            # execute notebook, transforming it in-place
-                            # TODO does it need to wiped first?
-                            if (
-                                "execution" in nb_bundle.nb.metadata
-                                and "timeout" in nb_bundle.nb.metadata.execution
-                            ):
-                                timeout = nb_bundle.nb.metadata.execution.timeout
-                            # TODO: for now we don't use the new record_timing feature
-                            # because it can affect test results
-                            executenb(
-                                nb_bundle.nb,
-                                cwd=tmpdirname,
-                                timeout=timeout,
-                                record_timing=False,
-                            )
-                    except Exception:
-                        exc_string = "".join(traceback.format_exc())
-                        self.logger.error("Execution Failed: {}".format(uri))
-                        yield create_bundle(
+                if in_temp:
+                    with tempfile.TemporaryDirectory() as tmpdirname:
+
+                        try:
+                            asset_files = _copy_assets(stage_record, tmpdirname)
+                        except Exception as err:
+                            yield ExecutionError("Assets Retrieval Error", uri, err)
+                            continue
+
+                        yield self.execute_single(
                             nb_bundle,
+                            uri,
                             tmpdirname,
+                            timeout,
+                            allow_errors,
                             asset_files,
-                            timer.last_split,
-                            exc_string,
-                        )
-                    else:
-                        self.logger.info("Execution Succeeded: {}".format(uri))
-                        yield create_bundle(
-                            nb_bundle, tmpdirname, asset_files, timer.last_split
                         )
+                else:
+                    yield self.execute_single(
+                        nb_bundle,
+                        uri,
+                        str(Path(uri).parent),
+                        timeout,
+                        allow_errors,
+                        None,
+                    )
+
             except Exception as err:
                 yield ExecutionError("Unexpected Error", uri, err)
+
+    def execute_single(self, nb_bundle, uri, cwd, timeout, allow_errors, asset_files):
+        result = single_nb_execution(
+            nb_bundle.nb, cwd=cwd, timeout=timeout, allow_errors=allow_errors,
+        )
+        if result.err:
+            self.logger.error("Execution Failed: {}".format(uri))
+            return _create_bundle(
+                nb_bundle, cwd, asset_files, result.time, result.exc_string,
+            )
+
+        self.logger.info("Execution Succeeded: {}".format(uri))
+        return _create_bundle(nb_bundle, cwd, asset_files, result.time, None)
+
+
+def _copy_assets(record, folder):
+    """Copy notebook assets to the folder the notebook will be executed in."""
+    asset_files = []
+    relative_paths = to_relative_paths(record.assets, Path(record.uri).parent)
+    for path, rel_path in zip(record.assets, relative_paths):
+        temp_file = Path(folder).joinpath(rel_path)
+        temp_file.parent.mkdir(parents=True, exist_ok=True)
+        shutil.copyfile(path, temp_file)
+        asset_files.append(temp_file)
+    return asset_files
+
+
+def _create_bundle(nb_bundle, tmpdirname, asset_files, exec_time, exec_tb):
+    """Create a cache bundle."""
+    return NbBundleIn(
+        nb_bundle.nb,
+        nb_bundle.uri,
+        # TODO retrieve assets that have changed file mtime?
+        artifacts=NbArtifacts(
+            [p for p in Path(tmpdirname).glob("**/*") if p not in asset_files],
+            tmpdirname,
+        )
+        if asset_files is not None
+        else None,
+        data={"execution_seconds": exec_time},
+        traceback=exec_tb,
+    )
